@@ -1,3 +1,4 @@
+
 package Data::TreeDumper ;
 
 use 5.006 ;
@@ -17,8 +18,8 @@ our %EXPORT_TAGS =
 
 our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } ) ;
 
-our @EXPORT = qw(DumpTree TreeDumper);
-our $VERSION = '0.07' ;
+our @EXPORT = qw(DumpTree TreeDumper CreateChainingFilter);
+our $VERSION = '0.08' ;
 
 use Term::Size;
 use Text::Wrap  ;
@@ -27,11 +28,13 @@ use Text::Wrap  ;
 # package variables  la Data::Dumper
 #----------------------------------------------------------------------
 
-our $Startlevel = 1 ;
-our $Useascii   = 1 ;
-our $Maxdepth   = -1 ;
-our $Filter     = undef ;
-our $Deparse    = 0 ; 
+our $Startlevel   = 1 ;
+our $Useascii     = 1 ;
+our $Maxdepth     = -1 ;
+our $Filter       = undef ;
+our $Virtualwidth = 120 ; 
+
+#~ our $Deparse    = 0 ;  # not implemented 
 
 sub DumpTree
 {
@@ -47,11 +50,12 @@ return
 			(
 			  $structure_to_dump
 			, {
-			    FILTER      => $Filter
-			  , START_LEVEL => $Startlevel
-			  , USE_ASCII   => $Useascii
-			  , MAX_DEPTH   => $Maxdepth
-			  , TITLE       => $title . "\n" 
+			    FILTER        => $Filter
+			  , START_LEVEL   => $Startlevel
+			  , USE_ASCII     => $Useascii
+			  , MAX_DEPTH     => $Maxdepth
+			  , TITLE         => $title . "\n" 
+			  , VIRTUAL_WIDTH => $Virtualwidth
 			  , %override
 			  }
 			)
@@ -68,47 +72,60 @@ my($class, @setup_data) = @_;
 
 my($this) = 
 	{
-	  FILTER      => undef
-	, START_LEVEL => 1
-	, USE_ASCII   => 1
-	, MAX_DEPTH   => -1
+	  FILTER        => undef
+	, START_LEVEL   => 1
+	, USE_ASCII     => 1
+	, MAX_DEPTH     => -1
+	, VIRTUAL_WIDTH => 120
 	, @setup_data
 	};
 
 return bless($this, $class);
 }
 
-sub Filter
+sub SetFilter
 {
 my($self, $filter) = @_;
 
 croak "Filter must be a code reference!" unless ('CODE' eq ref $filter) ;
 
-$self->{'filter'} = $filter;
+$self->{FILTER} = $filter;
 }
 
-sub StartLevel
+sub SetStartLevel
 {
 my($self, $start_level) = @_;
-$self->{'START_LEVEL'} = $start_level;
+$self->{START_LEVEL} = $start_level;
 }
 
 sub UseAscii
 {
 my($self, $use_ascii) = @_;
-$self->{'USE_ASCII'} = $use_ascii;
+$self->{USE_ASCII} = $use_ascii;
 }
 
 sub UseAnsi
 {
 my($self, $use_ansi) = @_;
-$self->{'USE_ASCII'} = (!$use_ansi) ;
+$self->{USE_ASCII} = (!$use_ansi) ;
 }
 
-sub Maxdepth
+sub SetMaxDepth
 {
 my($self, $max_depth) = @_;
-$self->{'MAX_DEPTH'} = $max_depth ;
+$self->{MAX_DEPTH} = $max_depth ;
+}
+
+sub SetIndentation
+{
+my($self, $indentation) = @_;
+$self->{INDENTATION} = $indentation ;
+}
+
+sub SetVirtualWidth
+{
+my($self, $width) = @_;
+$self->{VIRTUAL_WIDTH} = $width;
 }
 
 sub Dump
@@ -123,11 +140,12 @@ return
 			(
 			  $structure_to_dump
 			, {
-			    FILTER      => $self->{FILTER}
-			  , START_LEVEL => $self->{START_LEVEL}
-			  , USE_ASCII   => $self->{USE_ASCII}
-			  , MAX_DEPTH   => $self->{MAX_DEPTH}
-			  , TITLE       => $title . "\n"
+			    FILTER        => $self->{FILTER}
+			  , START_LEVEL   => $self->{START_LEVEL}
+			  , USE_ASCII     => $self->{USE_ASCII}
+			  , MAX_DEPTH     => $self->{MAX_DEPTH}
+			  , VIRTUAL_WIDTH => $self->{VIRTUAL_WIDTH}
+			  , TITLE         => $title . "\n"
 			  , %override
 			  }
 			)
@@ -145,60 +163,69 @@ my $setup            = shift ;
 my $level            = shift || 0 ;
 my $levels_left      = shift || [] ;
 
-my $filter_sub  = $setup->{FILTER} ;
-my $start_level = $setup->{START_LEVEL} ;
-my $use_ascii   = $setup->{USE_ASCII} ;
-my $max_depth   = $setup->{MAX_DEPTH} ;
-my $indentation = $setup->{INDENTATION} ;
+my $tree_type = ref $tree ;
+
+confess "TreeDumper can only display objects passed by reference!\n" unless(defined $tree_type) ;
+
+my $filter_sub    = $setup->{FILTER} ;
+my $start_level   = $setup->{START_LEVEL} ;
+my $use_ascii     = $setup->{USE_ASCII} ;
+my $max_depth     = $setup->{MAX_DEPTH} ;
+my $indentation   = $setup->{INDENTATION} ;
+my $virtual_width = $setup->{VIRTUAL_WIDTH} ;
 
 $start_level = 0   unless defined $start_level ;
 $use_ascii   = 1   unless defined $use_ascii ;
 $max_depth   = -1  unless defined $max_depth ;
 $indentation = ''  unless defined $indentation ;
+$virtual_width = 120 unless defined $virtual_width ;
+
+return('') if ($max_depth == $level) ;
 
 # used in the recursive call
 my $already_displayed_nodes = shift || {$tree => "ROOT 0", NEXT_INDEX => 1} ;
 
-my $tree_type = ref $tree ;
-
-confess "TreeDumper can only display objects passed by reference!\n" unless(defined $tree_type) ;
-
-return('') if ($max_depth == $level) ;
-
-my $output = '' ;
-my (@nodes_to_display, @node_names) ;
+# filters
+my ($replacement_tree, $nodes_to_display) ;
 
 if(defined $filter_sub)
 	{
-	($tree_type, @nodes_to_display) = $filter_sub->($tree, $level);
+	($tree_type, $replacement_tree, @$nodes_to_display) = $filter_sub->($tree, $level, $nodes_to_display) ;
+	$tree = $replacement_tree if(defined $replacement_tree) ;
 	}
 else
 	{
-	($tree_type, @nodes_to_display) = DefaultNodesToDisplay($tree, $level) ;
+	($tree_type, undef, @$nodes_to_display) = DefaultNodesToDisplay($tree, $level) ;
 	}
 	
 return('') unless defined $tree_type ;
 
-# Filters can change the name of the nodes by passing an array ref
+# filters can change the name of the nodes by passing an array ref
+my @node_names ;
+my @nodes_to_display = @$nodes_to_display ;
+
 for my $node (@nodes_to_display)
 	{
 	if('ARRAY' eq ref $node)
 		{
 		push @node_names, $node->[1] ;
-		$node = $node->[0] ; # Modify @nodes_to_display
+		$node = $node->[0] ; # Modify $nodes_to_display
 		}
 	else
 		{
 		push @node_names, $node ;
 		}
 	}
-	
+
+# wrapping	
 my ($columns, $rows) = Term::Size::chars *STDOUT{IO} ;
 
+$columns = $virtual_width if $columns eq '' ;
+
 local $Text::Wrap::columns  = $columns ;
 local $Text::Wrap::unexpand = 0 ;
-# need to set Text::Wrap::overflow type ?!?
 
+my $output = '' ;
 for (my $nodes_left = $#nodes_to_display ; $nodes_left >= 0 ; $nodes_left--)
 	{
 	my 
@@ -288,7 +315,7 @@ for (my $nodes_left = $#nodes_to_display ; $nodes_left >= 0 ; $nodes_left--)
 			$already_displayed_nodes->{NEXT_INDEX}++ ;
 			
 			my $address =  $already_displayed_nodes->{$element_ref} ;
-					
+			
 			$output .= wrap
 						(
 						  $tree_header
@@ -452,13 +479,24 @@ sub DefaultNodesToDisplay
 my $tree = shift ;
 
 my $tree_type = ref $tree ;
-# my $level = shift ; # not used by this sub
+my $level = shift ; # not used by this sub
+my $keys = shift ;
 
-return('HASH', sort keys %$tree)   if('HASH'    eq $tree_type) ;
-return('ARRAY', (0 .. @$tree - 1)) if('ARRAY'   eq $tree_type) ;
-return('REF', (0))                 if('REF'     eq $tree_type) ;
-return('CODE', (0))                if('CODE'    eq $tree_type) ;
-return('SCALAR', (0))              if('SCALAR'  eq $tree_type) ;
+if('HASH' eq $tree_type)
+	{
+	return('HASH', undef, sort keys %$tree) unless(defined $keys) ;
+	return('HASH', undef, @$keys) ;
+	}
+	
+if('ARRAY' eq $tree_type) 
+	{
+	return('ARRAY', undef, (0 .. @$tree - 1)) unless(defined $keys) ;
+	return('ARRAY', undef, @$keys) ;
+	}
+
+return('REF', undef, (0))                 if('REF'     eq $tree_type) ;
+return('CODE', undef, (0))                if('CODE'    eq $tree_type) ;
+return('SCALAR', undef, (0))              if('SCALAR'  eq $tree_type) ;
 
 my @nodes_to_display ;
 undef $tree_type ;
@@ -493,23 +531,66 @@ if($tree =~ /=/)
 		}
 	}
 
-return($tree_type, @nodes_to_display) ;
+return($tree_type, undef, @nodes_to_display) ;
 }
 
 #-------------------------------------------------------------------------------
 
 sub HashKeysSorter
 {
-# filter specialized in sorting hash keys
-
 my $structure_to_dump = shift ;
+my $level = shift ; # not used by this sub
+my $keys = shift ;
+
+if('HASH' eq ref $structure_to_dump)
+	{
+	return('HASH', undef, sort keys %$structure_to_dump) unless defined $keys ;
+	
+	my %keys ;
+	for my $key (@$keys)
+		{
+		if('ARRAY' eq ref $key)
+			{
+			$keys{$key->[0]} = $key ;
+			}
+		else
+			{
+			$keys{$key} = $key ;
+			}
+		}
+		
+	return('HASH', undef, map{$keys{$_}} sort keys %keys) ;
+	}
 
-return('HASH', sort keys %$structure_to_dump)   if('HASH'  eq $structure_to_dump) ;
 return(Data::TreeDumper::DefaultNodesToDisplay($structure_to_dump)) ;
 }
 
 #----------------------------------------------------------------------
 
+sub CreateChainingFilter
+{
+my @filters = @_ ;
+
+return sub
+	{
+	my $tree = shift ;
+	my $level = shift ;
+	my $keys = shift ;
+	
+	my ($tree_type, $replacement_tree);
+	
+	for my $filter (@filters)
+		{
+		($tree_type, $replacement_tree, @$keys) = $filter->($tree, $level, $keys) ;
+		$tree = $replacement_tree if (defined $replacement_tree) ;
+		}
+		
+	return ($tree_type, $replacement_tree, @$keys) ;
+	}
+} ;
+
+#----------------------------------------------------------------------
+
 sub GetSeparator 
 {
 # This sub is a good candidate for Memoize
@@ -593,8 +674,8 @@ Data::TreeDumper - dumps a data structure in a tree fashion.
   
   my $sub = sub {} ;
   
-  my $s = \%tree ;
-  (
+  my $s = 
+  {
   A => 
   	{
   	a => 
@@ -624,11 +705,11 @@ Data::TreeDumper - dumps a data structure in a tree fashion.
   		}
   	}
   , ARRAY => [qw(elment_1 element_2 element_3)]
-  ) ;
+  } ;
     
   
   #-------------------------------------------------------------------
-  # package global setup data
+  # package setup data
   #-------------------------------------------------------------------
   
   $Data::TreeDumper::Useascii = 0 ;
@@ -667,7 +748,7 @@ Data::TreeDumper - dumps a data structure in a tree fashion.
   
 =head1 Output
 
-  $tree:
+  title:
   |- A [H1]
   |  |- a [H2]
   |  |- bbbbbb = CODE(0x8139fa0) [C3]
@@ -688,8 +769,11 @@ Data::TreeDumper - dumps a data structure in a tree fashion.
 =head1 DESCRIPTION
 
 Data::Dumper and other modules do a great job at dumping data structure but their output sometime takes more
-brain to understand than it takes to understand the data itself. Data::TreeDumper dumps data in a
-trees like fashion I<hopping> the output will be easier on the beholder's eye but it might as well be the opposite.
+brain to understand than it takes to understand the data itself. When dumping big amounts of data, the output
+is overwelming and it's difficult to see the relationship between each piece of the dumped data.
+
+Data::TreeDumper dumps data in a trees like fashion I<hopping> for the output to be easier on the beholder's eye 
+and brain. But it might as well be the opposite!
 
 =head2 Address
 
@@ -747,29 +831,162 @@ Data::TreeDumper can sort the tree nodes with a user defined sub.
   FILTER => \&ReverseSort
   FILTER => \&Data::TreeDumper::HashKeysSorter
 
-The filter sub is passed two arguments, a reference to the node which is going to be displayed and
-it's depth (this allows you to selectively display elements at a certain depth). It returns the node's
-type and a list of 'keys' to display. The keys are hash keys or array indexes, you can thus reorder 
-both of them.
+The filter sub is passed three arguments, a reference to the node which is going to be displayed,
+it's depth (this allows you to selectively display elements at a certain depth) and an array reference
+containing the keys to be displayed (see filter chaining bellow) last argument can be undefined and can then
+be safely ignored.
 
-You can also remove entries by not returning their keys.
+a filter returns the node's type, an eventual new structure (see bellow) and a list of 'keys' to display.
+The keys are hash keys or array indexes.
+
+If you set FILTER to \&Data::TreeDumper::HashKeysSorter, hashes will be sorted in alphabetic order.
+
+=head3 Key removal
+
+Entries can be removed by not returning their keys.
 
   my $s = {visible => '', also_visible => '', not_visible => ''} ;
   my $OnlyVisible = sub
   	{
   	my $s = shift ;
   	
-	if('HASH' eq ref $sp)
+	if('HASH' eq ref $s)
   		{
-  		return('HASH', grep {! /^no_visible/} keys %$s) ;
+  		return('HASH', undef, grep {! /^not_visible/} keys %$s) ;
   		}
   		
   	return(Data::TreeDumper::DefaultNodesToDisplay($s)) ;
   	}
   	
   DumpTree($s, 'title', FILTER => $OnlyVisible) ;
+
+=head3 Label changing
+
+The label for a hash keys or an array index can be altered. This can be used to add visual information to the tree dump. Instead 
+for returning the key name, return an array reference containing the key name and the label you want to display.
+You only need to return such a reference for the entries you want to change thus a mix of scalars and array ref is acceptable.
+
+  sub StarOnA
+  {
+  # hash entries matching /^a/i have '*' prepended
   
-If you set FILTER to \&Data::TreeDumper::HashKeysSorter, hashes will be sorted in alphabetic order.
+  my $tree = shift ;
+  
+  if('HASH' eq ref $tree)
+  	{
+  	my @keys_to_dump ;
+  	
+  	for my $key_name (keys %$tree)
+  		{
+  		if($key_name =~ /^a/i)
+  			{
+  			$key_name = [$key_name, "* $key_name"] ;
+  			}
+  			
+  		push @keys_to_dump, $key_name ;
+  		}
+  		
+  	return ('HASH', undef, @keys_to_dump) ;
+  	}
+  	
+  return (Data::TreeDumper::DefaultNodesToDisplay($tree)) ;
+  }
+
+  print Data::TreeDumper::DumpTree($s, "Entries matching /^a/i have '*' prepended", FILTER => \&StarOnA) ;
+
+=head3 Structure replacement
+
+It is possible to replace the whole data structure in a filter. This comes handy when you want to display a 'worked'
+version of the structure. You can even change the type of the data structure, for example changing an array to a hash.
+
+  sub ReplaceArray
+  {
+  # replace arrays with hashes!!!
+  
+  my $tree = shift ;
+  
+  if('ARRAY' eq ref $tree)
+  	{
+	my $multiplication = $tree->[0] * $tree->[1] ;
+	my $replacement = {MULTIPLICATION => $multiplication} ;
+  	return('HASH', $replacement, keys %$replacement) ;
+  	}
+  	
+  return (Data::TreeDumper::DefaultNodesToDisplay($tree)) ;
+  }
+
+  print Data::TreeDumper::DumpTree($s, 'replace arrays with hashes!', FILTER => \&ReplaceArray) ;
+
+=head3 filter chaining
+
+It is possible to chain filters. B<Data::TreeDumper> exports I<CreateChainingFilter>. I<CreateChainingFilter>
+takes a list of filtering sub references. The filters must properly handle the third parameter passed to them.
+
+Suppose you want to chaine a filter, that adds a star before each hash key label, with a filter 
+that removes all (original) keys that match /^a/i.
+
+  sub AddStar
+  	{
+  	my $s = shift ;
+  	my $level = shift ;
+  	my $keys = shift ;
+  	
+  	if('HASH' eq ref $s)
+  		{
+  		$keys = [keys %$s] unless defined $keys ;
+  		
+  		my @new_keys ;
+  		
+  		for (@$keys)
+  			{
+  			if('' eq ref $_)
+  				{
+  				push @new_keys, [$_, "* $_"] ;
+  				}
+  			else
+  				{
+  				# another filter has changed the label
+  				push @new_keys, [$_->[0], "* $_->[1]"] ;
+  				}
+  			}
+  		
+  		return('HASH', undef, @new_keys) ;
+  		}
+  		
+  	return(Data::TreeDumper::DefaultNodesToDisplay($s)) ;
+  	} ;
+  	
+  sub RemoveA
+  	{
+  	my $s = shift ;
+  	my $level = shift ;
+  	my $keys = shift ;
+  	
+  	if('HASH' eq ref $s)
+  		{
+  		$keys = [keys %$s] unless defined $keys ;
+  		my @new_keys ;
+  		
+  		for (@$keys)
+  			{
+  			if('' eq ref $_)
+  				{
+  				push @new_keys, $_ unless /^a/i ;
+  				}
+  			else
+  				{
+  				# another filter has changed the label
+  				push @new_keys, $_ unless $_->[0] =~ /^a/i ;
+  				}
+  			}
+  		
+  		return('HASH', undef, @new_keys) ;
+  		}
+  		
+  	return(Data::TreeDumper::DefaultNodesToDisplay($s)) ;
+  	} ;
+  
+  DumpTree($s, 'Chained filters', FILTER => CreateChainingFilter(\&AddStar, \&RemoveA)) ;
 
 =head2 Start level
 
@@ -835,6 +1052,13 @@ wrapped multiple times so they snuggly fit your screen.
   |    Long_name Long_name Long_name Long_name Long_name Long_name
   |    Long_name Long_name Long_name Long_name Long_name [S26] = 0
 
+=head1 Zero width consol
+
+When no consol exists, while redirecting to a file for example, Data::TreeDumper uses the variable
+B<VIRTUAL_WIDTH> instead. Default is 120.
+
+	VIRTUAL_WIDTH => 120 ;
+
 =head1 Interfaces
 
 Data:TreeDumper has three interfaces. A 'package data' interface resembling Data::Dumper, an
@@ -844,10 +1068,11 @@ object oriented interface and the native interface. All interfaces return a stri
 
 =head3 Configuration Variables
 
-  $Data:TreeDumper::Startlevel = 1 ;
-  $Data:TreeDumper::Useascii   = 1 ;
-  $Data:TreeDumper::Maxdepth   = -1 ;
-  $Data:TreeDumper::Filter     = \&FlipEverySecondOne ;
+  $Data:TreeDumper::Startlevel   = 1 ;
+  $Data:TreeDumper::Useascii     = 1 ;
+  $Data:TreeDumper::Maxdepth     = -1 ;
+  $Data:TreeDumper::Virtualwidth = 120 ;
+  $Data:TreeDumper::Filter       = \&FlipEverySecondOne ;
   
 =head3 Function
 
@@ -894,21 +1119,21 @@ B<DumpTree> uses the configuration variables defined above. It takes the followi
   
 =head1 Bugs
 
-Text::Wrap sometimes dies when the output of a script using B<Data::TreeDumper> is
-redirected into a file.
-
-Plenty more , lurking in the dark corners, waiting to be found.
+None I know of in this release but plenty, lurking in the dark corners, waiting to be found.
 
 =head1 Examples
 
-Two examples files are included in the distribution.
+Three examples files are included in the distribution.
 
 I<usage.pl> shows you how you can use B<Data::TreeDumper>.
+
+I<filters.pl> shows you how you how to do advance filtering.
+
 I<try_it.pl> is meant as a scratch pad for you to try B<Data::TreeDumper>.
 
 =head1 EXPORT
 
-I<DumpTree> and I<TreeDumper>.
+I<DumpTree>, I<TreeDumper> and  I<CreateChainingFilter>.
 
 =head1 AUTHOR
 
